@@ -18,7 +18,7 @@
  *
  * The author may be contacted via http://dmalloc.com/
  *
- * $Id: chunk.c,v 1.195 2003/09/09 23:18:09 gray Exp $
+ * $Id: chunk.c,v 1.201 2004/01/28 14:29:43 gray Exp $
  */
 
 /*
@@ -91,6 +91,9 @@ static char *information = "@(#) $Information: lock-threads is enabled $";
 /* limit in how much memory we are allowed to allocate */
 unsigned long		_dmalloc_memory_limit = 0;
 
+/* total number of bytes that the heap has allocated */
+unsigned long		_dmalloc_alloc_total = 0;
+
 /*
  * local variables
  */
@@ -136,14 +139,13 @@ static	unsigned long	alloc_current = 0;	/* current memory usage */
 static	unsigned long	alloc_maximum = 0;	/* maximum memory usage  */
 static	unsigned long	alloc_cur_given = 0;	/* current mem given */
 static	unsigned long	alloc_max_given = 0;	/* maximum mem given  */
-static	unsigned long	alloc_total = 0;	/* total allocation */
 static	unsigned long	alloc_one_max = 0;	/* maximum at once */
 static	unsigned long	free_space_bytes = 0;	/* count the free bytes */
 
 /* pointer stats */
 static	unsigned long	alloc_cur_pnts = 0;	/* current pointers */
 static	unsigned long	alloc_max_pnts = 0;	/* maximum pointers */
-static	unsigned long	alloc_tot_pnts = 0;	/* current pointers */
+static	unsigned long	alloc_tot_pnts = 0;	/* total pointers */
 
 /* admin counts */
 static	unsigned long	heap_check_c = 0;	/* count of heap-checks */
@@ -1632,11 +1634,13 @@ static	int	check_used_slot(const skip_alloc_t *slot_p,
     return 0;
   }
   
-  /* check our size */
+#if LARGEST_ALLOCATION
+  /* have we exceeded the upper bounds */
   if (slot_p->sa_user_size > LARGEST_ALLOCATION) {
     dmalloc_errno = ERROR_BAD_SIZE;
     return 0;
   }
+#endif
   
   /* check our total block size */
   if (slot_p->sa_total_size > BLOCK_SIZE / 2
@@ -2216,9 +2220,16 @@ int	_dmalloc_chunk_pnt_check(const char *func, const void *user_pnt,
 {
   skip_alloc_t	*slot_p;
   unsigned int	min;
+  pnt_info_t	pnt_info;
   
   if (BIT_IS_SET(_dmalloc_flags, DEBUG_LOG_TRANS)) {
-    dmalloc_message("checking pointer '%#lx'", (unsigned long)user_pnt);
+    if (func == NULL) {
+      dmalloc_message("checking pointer '%#lx'", (unsigned long)user_pnt);
+    }
+    else {
+      dmalloc_message("checking func '%s' pointer '%#lx'",
+		      func, (unsigned long)user_pnt);
+    }
   }
   
   /* try to find the address */
@@ -2251,16 +2262,16 @@ int	_dmalloc_chunk_pnt_check(const char *func, const void *user_pnt,
       min = min_size;
     }
     else {
+      /* length of the string + 1 for the \0 */
       min = strlen((char *)user_pnt) + 1;
     }
     
-    if (BIT_IS_SET(slot_p->sa_flags, ALLOC_FLAG_FENCE)) {
-      min += FENCE_OVERHEAD_SIZE;
-    }
+    /* get info about the pointer */
+    get_pnt_info(slot_p, &pnt_info);
     
-    /* do we overflow the memory slot */
-    if ((char *)user_pnt + min >
-	(char *)slot_p->sa_mem + slot_p->sa_user_size) {
+    /* are we within bounds on the user pointer and size */
+    if ((char *)user_pnt < (char *)pnt_info.pi_user_start
+	|| (char *)user_pnt + min > (char *)pnt_info.pi_user_bounds) {
       dmalloc_errno = ERROR_WOULD_OVERWRITE;
       log_error_info(NULL, 0, slot_p->sa_file, slot_p->sa_line,
 		     user_pnt, 0, NULL, "pointer-check");
@@ -2341,6 +2352,7 @@ void	*_dmalloc_chunk_malloc(const char *file, const unsigned int line,
   }
 #endif
   
+#if LARGEST_ALLOCATION
   /* have we exceeded the upper bounds */
   if (size > LARGEST_ALLOCATION) {
     dmalloc_errno = ERROR_TOO_BIG;
@@ -2348,6 +2360,7 @@ void	*_dmalloc_chunk_malloc(const char *file, const unsigned int line,
     dmalloc_error("_dmalloc_chunk_malloc");
     return MALLOC_ERROR;
   }
+#endif
   
   needed_size = size;
   
@@ -2453,7 +2466,7 @@ void	*_dmalloc_chunk_malloc(const char *file, const unsigned int line,
   /* monitor current allocation level */
   alloc_current += size;
   alloc_maximum = MAX(alloc_maximum, alloc_current);
-  alloc_total += size;
+  _dmalloc_alloc_total += size;
   alloc_one_max = MAX(alloc_one_max, size);
   
   /* monitor pointer usage */
@@ -2548,7 +2561,8 @@ int	_dmalloc_chunk_free(const char *file, const unsigned int line,
   
   if (! check_used_slot(slot_p, user_pnt, 1 /* exact pnt */)) {
     /* error set in check slot */
-    log_error_info(file, line, NULL, 0, user_pnt, 0, NULL, "free");
+    log_error_info(file, line, slot_p->sa_file, slot_p->sa_line, user_pnt,
+		   0, NULL, "free");
     dmalloc_error("_dmalloc_chunk_free");
     return FREE_ERROR;
   }
@@ -2701,7 +2715,7 @@ void	*_dmalloc_chunk_realloc(const char *file, const unsigned int line,
   slot_p = find_address(old_user_pnt, 0 /* not exact pointer */, skip_update);
   if (slot_p == NULL) {
     dmalloc_errno = ERROR_NOT_FOUND;
-    log_error_info(NULL, 0, NULL, 0, old_user_pnt, 0, NULL,
+    log_error_info(file, line, NULL, 0, old_user_pnt, 0, NULL,
 		   "_dmalloc_chunk_realloc");
     dmalloc_error("_dmalloc_chunk_realloc");
     return 0;
@@ -2756,7 +2770,7 @@ void	*_dmalloc_chunk_realloc(const char *file, const unsigned int line,
      */
     alloc_current += new_size - old_size;
     alloc_maximum = MAX(alloc_maximum, alloc_current);
-    alloc_total += new_size;
+    _dmalloc_alloc_total += new_size;
     alloc_one_max = MAX(alloc_one_max, new_size);
     
     /* monitor pointer usage */
@@ -2874,7 +2888,7 @@ void	_dmalloc_chunk_log_stats(void)
   dmalloc_message("  current memory in use: %lu bytes (%lu pnts)",
 		  alloc_current, alloc_cur_pnts);
   dmalloc_message(" total memory allocated: %lu bytes (%lu pnts)",
-		  alloc_total, alloc_tot_pnts);
+		  _dmalloc_alloc_total, alloc_tot_pnts);
   
   /* maximum stats */
   dmalloc_message(" max in use at one time: %lu bytes (%lu pnts)",
@@ -2900,8 +2914,7 @@ void	_dmalloc_chunk_log_stats(void)
  *
  * DESCRIPTION:
  *
- * Dump the pointer information that has changed since a pointer in
- * time.
+ * Log the pointers that has changed since a pointer in time.
  *
  * RETURNS:
  *
@@ -3062,3 +3075,87 @@ void	_dmalloc_chunk_log_changed(const unsigned long mark,
     }
   }
 }
+
+/*
+ * unsigned long _dmalloc_chunk_count_changed
+ *
+ * DESCRIPTION:
+ *
+ * Return the pointers that has changed since a pointer in time.
+ *
+ * RETURNS:
+ *
+ * Number of bytes changed since mark.
+ *
+ * ARGUMENTS:
+ *
+ * mark -> Dmalloc counter used to mark a specific time so that
+ * servers can check on the changed pointers.
+ *
+ * count_non_free_b -> If set to 1 then count the new not-freed
+ * (i.e. used) pointers.
+ *
+ * count_free_b -> If set to 1 then count the new freed pointers.
+ */
+unsigned long	_dmalloc_chunk_count_changed(const unsigned long mark,
+					     const int count_not_freed_b,
+					     const int count_freed_b)
+{
+  skip_alloc_t	*slot_p;
+  int		freed_b, used_b;
+  int		checking_list_c = 0;
+  unsigned int	mem_count = 0;
+  
+  /* run through the blocks */
+  for (slot_p = skip_address_list->sa_next_p[0];
+       ;
+       slot_p = slot_p->sa_next_p[0]) {
+    
+    /*
+     * switch to the free list in the middle after we've checked the
+     * used pointer slots
+     */
+    if (slot_p == NULL) {
+      checking_list_c++;
+      if (checking_list_c == 1) {
+	slot_p = skip_free_list->sa_next_p[0];
+      }
+#if FREED_POINTER_DELAY
+      else if (checking_list_c == 2) {
+	slot_p = free_wait_list_head;
+      }
+#endif
+      else {
+	/* we are done */
+	break;
+      }
+      if (slot_p == NULL) {
+	break;
+      }
+    }
+    
+    freed_b = BIT_IS_SET(slot_p->sa_flags, ALLOC_FLAG_FREE);
+    used_b = BIT_IS_SET(slot_p->sa_flags, ALLOC_FLAG_USER);
+    
+    /*
+     * check for different types
+     */
+    if (! (freed_b || used_b)) {
+      continue;
+    }
+    /* is it too long ago? */
+    if (slot_p->sa_use_iter <= mark) {
+      continue;
+    }
+    
+    /* count the memory */
+    if (count_not_freed_b && used_b) {
+      mem_count += slot_p->sa_user_size;
+    }
+    else if (count_freed_b && freed_b) {
+      mem_count += slot_p->sa_user_size;
+    }
+  }
+  
+  return mem_count;
+}
