@@ -18,7 +18,7 @@
  *
  * The author may be contacted via http://dmalloc.com/
  *
- * $Id: dmalloc_t.c,v 1.106 2003/09/08 15:21:41 gray Exp $
+ * $Id: dmalloc_t.c,v 1.112 2004/01/28 14:30:59 gray Exp $
  */
 
 /*
@@ -1087,7 +1087,313 @@ static	int	check_special(void)
     
     dmalloc_debug(old_flags);
   }
- 
+  
+  /********************/
+  
+  /*
+   * Make sure that the start file:line works.
+   */
+  {
+    unsigned int	old_flags = dmalloc_debug_current();
+    int			our_errno_hold = dmalloc_errno;
+    char		*loc_file, save_ch;
+    int			iter_c, loc_line;
+    void		*pnts[2];
+    char		setup[128];
+    
+    /* turn on fence post checking */
+    dmalloc_debug(DEBUG_CHECK_FENCE);
+    dmalloc_errno = 0;
+    
+    if (! silent_b) {
+      (void)printf("  Checking heap check start at file:line\n");
+    }
+    
+#define BUF_SIZE	64
+    
+    /* make an allocation */
+    pnts[0] = malloc(BUF_SIZE);
+    if (pnts[0] == NULL) {
+      if (! silent_b) {
+	(void)printf("   ERROR: could not malloc %d bytes.\n", BLOCK_SIZE);
+      }
+      return 0;
+    }
+    
+    /* save the character but then overwrite the high fence post */
+    save_ch = *((char *)pnts[0] + BUF_SIZE);
+    *((char *)pnts[0] + BUF_SIZE) = '\0';
+    
+    /* make another allocation */
+    pnts[1] = malloc(BUF_SIZE);
+    if (pnts[1] == NULL) {
+      if (! silent_b) {
+	(void)printf("   ERROR: could not malloc %d bytes.\n", BLOCK_SIZE);
+      }
+      return 0;
+    }
+    
+    /* it shouldn't generate an error */
+    if (dmalloc_errno != 0) {
+      if (! silent_b) {
+	(void)printf("   ERROR: should not have gotten an error with no heap checking enabled.\n");
+      }
+      return 0;
+    }
+    
+    /* restore the overwritten character otherwise we can't free the pointer */
+    *((char *)pnts[0] + BUF_SIZE) = save_ch;
+    free(pnts[0]);
+    free(pnts[1]);
+    
+    for (iter_c = 0; iter_c < 2; iter_c++) {
+      /*
+       * we have to do this loop hack here because we need to know the
+       * __FILE__ and __LINE__ of a certain location to set the
+       * variable so then we need to run it again.
+       */
+      
+      /*
+       * Make an allocation recording where we did it.
+       *
+       * NOTE: This all needs to be on the same line.
+       */
+      loc_file= __FILE__; loc_line = __LINE__; pnts[iter_c] = malloc(BUF_SIZE);
+      if (pnts[iter_c] == NULL) {
+	if (! silent_b) {
+	  (void)printf("   ERROR: could not malloc %d bytes.\n", BLOCK_SIZE);
+	}
+	return 0;
+      }
+      
+      /* first time through the loop? */
+      if (iter_c == 0) {
+	/* save the character and then overwrite the high fence post */
+	save_ch = *((char *)pnts[0] + BUF_SIZE);
+	*((char *)pnts[0] + BUF_SIZE) = '\0';
+	
+	/*
+	 * build and enable an options string turning on checking at
+	 * the above allocation
+	 */
+	(void)sprintf(setup, "debug=%#x,start=%s:%d",
+		      DEBUG_CHECK_FENCE, loc_file, loc_line);
+	dmalloc_debug_setup(setup);	
+	continue;
+      }
+      
+      /*
+       * now the 2nd time through the loop we should have seen an
+       * error because heap checking should have been enabled at the
+       * 2nd allocation so the heap should have been checked and the
+       * problem with the 1st allocation detected.
+       */
+      if (dmalloc_errno != ERROR_OVER_FENCE) {
+	if (! silent_b) {
+	  (void)printf("   ERROR: should have gotten over fence-post error after checking started.\n");
+	}
+	return 0;
+      }
+      
+      /*
+       * restore the overwritten character otherwise we can't free
+       * the pointer
+       */
+      *((char *)pnts[0] + BUF_SIZE) = save_ch;
+    }
+    
+    free(pnts[0]);
+    free(pnts[1]);
+    
+    /* reset the debug flags and errno */
+    dmalloc_debug(old_flags);
+    dmalloc_errno = our_errno_hold;
+  }
+  
+  /********************/
+  
+  /*
+   * Make sure that the start iteration count works.
+   */
+  {
+    unsigned int	old_flags = dmalloc_debug_current();
+    int			our_errno_hold = dmalloc_errno;
+    char		save_ch;
+    void		*pnt2;
+    char		setup[128];
+    
+    /* turn on fence post checking */
+    dmalloc_debug(DEBUG_CHECK_FENCE);
+    dmalloc_errno = 0;
+    
+    if (! silent_b) {
+      (void)printf("  Checking heap check start at iteration count\n");
+    }
+    
+#define BUF_SIZE	64
+    
+    /* make an allocation */
+    pnt = malloc(BUF_SIZE);
+    if (pnt == NULL) {
+      if (! silent_b) {
+	(void)printf("   ERROR: could not malloc %d bytes.\n", BLOCK_SIZE);
+      }
+      return 0;
+    }
+    
+    /* overwrite the high fence post */
+    save_ch = *((char *)pnt + BUF_SIZE);
+    *((char *)pnt + BUF_SIZE) = '\0';
+    
+    /* make another allocation */
+    pnt2 = malloc(BUF_SIZE);
+    if (pnt2 == NULL) {
+      if (! silent_b) {
+	(void)printf("   ERROR: could not malloc %d bytes.\n", BLOCK_SIZE);
+      }
+      return 0;
+    }
+    
+    /* it shouldn't generate an error */
+    if (dmalloc_errno != 0) {
+      if (! silent_b) {
+	(void)printf("   ERROR: should not have gotten an error with no heap checking enabled.\n");
+      }
+      return 0;
+    }
+    
+    /* free the 2nd pointer */
+    free(pnt2);
+    
+    /*
+     * build and enable an options string turning on checking at the
+     * next transaction
+     */
+    (void)sprintf(setup, "debug=%#x,start=c1", DEBUG_CHECK_FENCE);
+    dmalloc_debug_setup(setup);	
+    
+    /*
+     * make another allocation which should enable heap checking and
+     * notice the above pointer overwrite
+     */
+    pnt2 = malloc(BUF_SIZE);
+    if (pnt2 == NULL) {
+      if (! silent_b) {
+	(void)printf("   ERROR: could not malloc %d bytes.\n", BLOCK_SIZE);
+      }
+      return 0;
+    }
+    
+    /* now we should see the error */
+    if (dmalloc_errno != ERROR_OVER_FENCE) {
+      if (! silent_b) {
+	(void)printf("   ERROR: should have gotten over fence-post error after checking started.\n");
+      }
+      return 0;
+    }
+    
+    /* restore the overwritten character otherwise we can't free the pointer */
+    *((char *)pnt + BUF_SIZE) = save_ch;
+    free(pnt);
+    free(pnt2);
+    
+    /* reset the debug flags and errno */
+    dmalloc_debug(old_flags);
+    dmalloc_errno = our_errno_hold;
+  }
+  
+  /********************/
+  
+  /*
+   * Make sure that the start after memory size allocated.
+   */
+  {
+    unsigned int	old_flags = dmalloc_debug_current();
+    int			our_errno_hold = dmalloc_errno;
+    char		save_ch;
+    void		*pnt2;
+    char		setup[128];
+    
+    /* turn on fence post checking */
+    dmalloc_debug(DEBUG_CHECK_FENCE);
+    dmalloc_errno = 0;
+    
+    if (! silent_b) {
+      (void)printf("  Checking heap check start at memory size\n");
+    }
+    
+#define BUF_SIZE	64
+    
+    /* make an allocation */
+    pnt = malloc(BUF_SIZE);
+    if (pnt == NULL) {
+      if (! silent_b) {
+	(void)printf("   ERROR: could not malloc %d bytes.\n", BLOCK_SIZE);
+      }
+      return 0;
+    }
+    
+    /* overwrite the high fence post */
+    save_ch = *((char *)pnt + BUF_SIZE);
+    *((char *)pnt + BUF_SIZE) = '\0';
+    
+    /* make another allocation */
+    pnt2 = malloc(BUF_SIZE);
+    if (pnt2 == NULL) {
+      if (! silent_b) {
+	(void)printf("   ERROR: could not malloc %d bytes.\n", BLOCK_SIZE);
+      }
+      return 0;
+    }
+    
+    /* it shouldn't generate an error */
+    if (dmalloc_errno != 0) {
+      if (! silent_b) {
+	(void)printf("   ERROR: should not have gotten an error with no heap checking enabled.\n");
+      }
+      return 0;
+    }
+    
+    free(pnt2);
+    
+    /*
+     * build and enable an options string turning on checking at the
+     * next transaction
+     */
+    (void)sprintf(setup, "debug=%#x,start=s%lu",
+		  DEBUG_CHECK_FENCE, dmalloc_memory_allocated());
+    dmalloc_debug_setup(setup);	
+    
+    /*
+     * make another allocation which should enable heap checking and
+     * notice the above pointer overwrite
+     */
+    pnt2 = malloc(BUF_SIZE);
+    if (pnt2 == NULL) {
+      if (! silent_b) {
+	(void)printf("   ERROR: could not malloc %d bytes.\n", BLOCK_SIZE);
+      }
+      return 0;
+    }
+    
+    /* now we should see the error */
+    if (dmalloc_errno != ERROR_OVER_FENCE) {
+      if (! silent_b) {
+	(void)printf("   ERROR: should have gotten over fence-post error after checking started.\n");
+      }
+      return 0;
+    }
+    
+    /* restore the overwritten character otherwise we can't free the pointer */
+    *((char *)pnt + BUF_SIZE) = save_ch;
+    free(pnt);
+    free(pnt2);
+    
+    /* reset the debug flags and errno */
+    dmalloc_debug(old_flags);
+    dmalloc_errno = our_errno_hold;
+  }
+  
   /********************/
  
   dmalloc_message("NOTE: ignore the errors from the above ----- to here.\n");
@@ -1471,7 +1777,219 @@ static	int	check_special(void)
   
   /********************/
   
-  /* NOTE: add tests which get errors before the ------- message above */
+  /*
+   * Make sure that the calloc macro is valid.
+   */
+  {
+    DMALLOC_SIZE	user_size;
+  
+    if (! silent_b) {
+      (void)printf("  Checking calloc macro arguments\n");
+    }
+    
+    /* notice that we do not have a () around this operation */
+#define SIZE_ARG	10 + 1
+#define COUNT_ARG	2
+    
+    /*
+     * Initially I was not putting parens around macro arguments in
+     * dmalloc.h.  Rookie mistake.  This should allocate (10 + 1) * 2
+     * bytes (22) not 10 + 1 * 2 bytes (12).
+     */
+    pnt = calloc(SIZE_ARG, COUNT_ARG);
+    if (pnt == NULL) {
+      if (! silent_b) {
+	(void)printf("   ERROR: could not calloc %d bytes.\n",
+		     (SIZE_ARG) * (COUNT_ARG));
+      }
+      final = 0;
+    }
+    
+    else if (dmalloc_examine(pnt, &user_size, NULL, NULL, NULL, NULL, NULL,
+			     NULL) != DMALLOC_NOERROR) {
+      if (! silent_b) {
+	(void)printf("   ERROR: examining pointer %lx failed.\n",
+		     (unsigned long)pnt);
+      }
+      final = 0;
+    }
+    
+    else if (user_size != (SIZE_ARG) * (COUNT_ARG)) {
+      if (! silent_b) {
+	(void)printf("   ERROR: calloc size should be %d but was %d.\n",
+		     (SIZE_ARG) * (COUNT_ARG), user_size);
+      }
+      final = 0;
+    }
+  }
+  
+  /********************/
+  
+  /*
+   * Verify that the check-funcs work with check-fence.  Thanks to
+   * John Hetherington for reporting this.
+   */
+  {
+    unsigned int	old_flags = dmalloc_debug_current();
+    int			our_errno_hold = dmalloc_errno;
+    
+    dmalloc_debug(DEBUG_CHECK_FUNCS);
+    
+    if (! silent_b) {
+      (void)printf("  Checking check-funcs with check-fence\n");
+    }
+    
+#define BUF_SIZE	64
+    
+    pnt = malloc(BUF_SIZE);
+    if (pnt == NULL) {
+      if (! silent_b) {
+	(void)printf("   ERROR: could not malloc %d bytes.\n", BLOCK_SIZE);
+      }
+      return 0;
+    }
+    
+    dmalloc_errno = ERROR_NONE;
+    _dmalloc_memset(pnt, 0, BUF_SIZE);
+    if (dmalloc_errno != ERROR_NONE) {
+      if (! silent_b) {
+	(void)printf("   ERROR: memset on buf of %d bytes failed.\n",
+		     BUF_SIZE);
+      }
+      final = 0;
+    }
+    
+    free(pnt);
+    
+    /* now turn on the check-fence token */
+    dmalloc_debug(DEBUG_CHECK_FENCE | DEBUG_CHECK_FUNCS);
+    
+    pnt = malloc(BUF_SIZE);
+    if (pnt == NULL) {
+      if (! silent_b) {
+	(void)printf("   ERROR: could not malloc %d bytes.\n", BLOCK_SIZE);
+      }
+      return 0;
+    }
+    
+    dmalloc_errno = ERROR_NONE;
+    _dmalloc_memset(pnt, 0, BUF_SIZE);
+    if (dmalloc_errno != ERROR_NONE) {
+      if (! silent_b) {
+	(void)printf("   ERROR: memset of %d bytes with check-fence failed.\n",
+		     BUF_SIZE);
+      }
+      final = 0;
+    }
+    
+    free(pnt);
+    dmalloc_debug(old_flags);
+    
+    /* recover the errno if necessary */
+    if (dmalloc_errno == ERROR_NONE) {
+      dmalloc_errno = our_errno_hold;
+    }
+  }
+  
+  /********************/
+  
+  /*
+   * Verify the dmalloc_count_changed function.
+   */
+  {
+    unsigned long	size, mem_count, loc_mark;
+    
+    if (! silent_b) {
+      (void)printf("  Checking dmalloc_count_changed function\n");
+    }
+    
+    /* get a mark */
+    loc_mark = dmalloc_mark();
+    
+    /* make sure that nothing has changed */
+    mem_count = dmalloc_count_changed(loc_mark, 1 /* not-freed */,
+				      1 /* freed */);
+    if (mem_count != 0) {
+      if (! silent_b) {
+	(void)printf("   ERROR: count-changed reported %lu bytes changed.\n",
+		     mem_count);
+      }
+      return 0;
+    }
+    
+    /* allocate a pointer */
+    size = _dmalloc_rand() % MAX_ALLOC + 10;
+    pnt = malloc(size);
+    if (pnt == NULL) {
+      if (! silent_b) {
+	(void)printf("   ERROR: could not malloc %lu bytes.\n", size);
+      }
+      return 0;
+    }
+    
+    /* make sure that we see the non-freed pointer */
+    mem_count = dmalloc_count_changed(loc_mark, 1 /* not-freed */,
+				      0 /* no freed */);
+    if (mem_count != size) {
+      if (! silent_b) {
+	(void)printf("   ERROR: count-changed reported %lu bytes changed not %lu.\n",
+		     mem_count, size);
+      }
+      return 0;
+    }
+    /* make sure that we see non-freed pointer with non-fred + freed flags */
+    mem_count = dmalloc_count_changed(loc_mark, 1 /* not-freed */,
+				      1 /* no freed */);
+    if (mem_count != size) {
+      if (! silent_b) {
+	(void)printf("   ERROR: count-changed reported %lu bytes changed not %lu.\n",
+		     mem_count, size);
+      }
+      return 0;
+    }
+    
+    /* now free it */
+    free(pnt);
+    
+    /* make sure that everything is freed */
+    mem_count = dmalloc_count_changed(loc_mark, 1 /* not-freed */,
+				      0 /* no freed */);
+    if (mem_count != 0) {
+      if (! silent_b) {
+	(void)printf("   ERROR: count-changed reported %lu bytes changed.\n",
+		     mem_count);
+      }
+      return 0;
+    }
+    
+    /* make sure that we see the freed pointer */
+    mem_count = dmalloc_count_changed(loc_mark, 0 /* no not-freed */,
+				      1 /* no freed */);
+    if (mem_count != size) {
+      if (! silent_b) {
+	(void)printf("   ERROR: count-changed report %lu bytes changed not %lu.\n",
+		     mem_count, size);
+      }
+      return 0;
+    }
+    /* make sure that we see the freed pointer with both flags */
+    mem_count = dmalloc_count_changed(loc_mark, 1 /* no not-freed */,
+				      1 /* no freed */);
+    if (mem_count != size) {
+      if (! silent_b) {
+	(void)printf("   ERROR: count-changed report %lu bytes changed not %lu.\n",
+		     mem_count, size);
+      }
+      return 0;
+    }
+  }
+ 
+  /********************/
+  
+  /*
+   * NOTE: add tests which should result in errors before the -------
+   * message above
+   */
   
   return final;
 }
@@ -1509,12 +2027,12 @@ static	void	do_interactive(void)
       (void)printf("\tmalloc    - allocate memory\n");
       (void)printf("\tcalloc    - allocate/clear memory\n");
       (void)printf("\trealloc   - reallocate memory\n");
+      (void)printf("\trecalloc  - reallocate cleared memory\n");
       (void)printf("\tmemalign  - allocate aligned memory\n");
       (void)printf("\tvalloc    - allocate page-aligned memory\n");
       (void)printf("\tstrdup    - allocate a string\n");
       (void)printf("\tfree      - deallocate memory\n\n");
       
-      (void)printf("\tmap       - map the heap to the logfile\n");
       (void)printf("\tstats     - dump heap stats to the logfile\n");
       (void)printf("\tunfreed   - list the unfree memory to the logfile\n");
       (void)printf("\tmark      - display the current mark value\n");
