@@ -18,7 +18,7 @@
  *
  * The author may be contacted via http://dmalloc.com/
  *
- * $Id: malloc.c,v 1.175 2003/09/29 22:58:27 gray Exp $
+ * $Id: malloc.c,v 1.178 2004/01/28 14:30:40 gray Exp $
  */
 
 /*
@@ -122,7 +122,8 @@ static	dmalloc_track_t	tracking_func = NULL;	/* memory trxn tracking func */
 /* debug variables */
 static	char		*start_file = NULL;	/* file to start at */
 static	int		start_line = 0;		/* line to start */
-static	int		start_count = 0;	/* start after X */
+static	unsigned long	start_iter = 0;		/* start after X iterations */
+static	unsigned long	start_size = 0;		/* start after X bytes */
 static	int		thread_lock_c = 0;	/* lock counter */
 
 /****************************** thread locking *******************************/
@@ -286,11 +287,11 @@ static	void	process_environ(const char *option_str)
 			   (long *)&_dmalloc_address_seen_n, &_dmalloc_flags,
 			   &_dmalloc_check_interval, &_dmalloc_lock_on,
 			   &dmalloc_logpath, &start_file, &start_line,
-			   &start_count, &_dmalloc_memory_limit);
+			   &start_iter, &start_size, &_dmalloc_memory_limit);
   thread_lock_c = _dmalloc_lock_on;
   
   /* if we set the start stuff, then check-heap comes on later */
-  if (start_count > 0) {
+  if (start_iter > 0 || start_size > 0) {
     BIT_CLEAR(_dmalloc_flags, DEBUG_CHECK_HEAP);
   }
   
@@ -382,7 +383,8 @@ static	int	dmalloc_startup(const char *debug_str)
      * flag is cleared.
      */ 
     if (start_file != NULL
-	|| start_count > 0
+	|| start_iter > 0
+	|| start_size > 0
 	|| _dmalloc_check_interval > 0) {
       BIT_CLEAR(_dmalloc_flags, DEBUG_CHECK_HEAP);
     }
@@ -456,10 +458,28 @@ static	int	dmalloc_startup(const char *debug_str)
 }
 
 /*
+ * static int dmalloc_in
+ *
+ * DESCRIPTION:
+ *
  * Call to the alloc routines has been made.  Do some initialization,
  * locking, and check some debug variables.
  *
- * Returns ERROR or NOERROR.
+ * RETURNS:
+ *
+ * Success - 1
+ *
+ * Failure - 0
+ *
+ * ARGUMENTS:
+ *
+ * file -> File-name or return-address of the caller.
+ *
+ * line -> Line-number of the caller.
+ *
+ * check_heap_b -> Set to 1 if it is okay to check the heap.  If set
+ * to 0 then the caller will check it itself or it is a non-invasive
+ * call.
  */
 static	int	dmalloc_in(const char *file, const int line,
 			   const int check_heap_b)
@@ -512,12 +532,22 @@ static	int	dmalloc_in(const char *file, const int line,
   }
   
   /* start checking heap after X times */
-  else if (start_count > 0) {
-    if (--start_count == 0) {
+  else if (start_iter > 0) {
+    if (--start_iter == 0) {
       BIT_SET(_dmalloc_flags, DEBUG_CHECK_HEAP);
+      /*
+       * this is automatically disabled since it goes to 0 so the
+       * interval can go on/off
+       */
     }
   }
   
+  else if (start_size > 0 && start_size >= _dmalloc_alloc_total) {
+    BIT_SET(_dmalloc_flags, DEBUG_CHECK_HEAP);
+    start_size = 0;
+    /* disable this check so the interval can go on/off */
+  }
+  
   /* checking heap every X times */
   else if (_dmalloc_check_interval > 0) {
     if (_dmalloc_iter_c % _dmalloc_check_interval == 0) {
@@ -610,7 +640,8 @@ void	dmalloc_shutdown(void)
    * not been overwritten.  Thanks Randell.
    */
   if (BIT_IS_SET(_dmalloc_flags, DEBUG_CHECK_HEAP)
-      || BIT_IS_SET(_dmalloc_flags, DEBUG_CHECK_BLANK)) {
+      || BIT_IS_SET(_dmalloc_flags, DEBUG_CHECK_BLANK)
+      || BIT_IS_SET(_dmalloc_flags, DEBUG_CHECK_SHUTDOWN)) {
     (void)_dmalloc_chunk_heap_check();
   }
   
@@ -1353,6 +1384,64 @@ int	malloc_verify(const DMALLOC_PNT pnt)
   return dmalloc_verify(pnt);
 }
 
+/*
+ * int dmalloc_verify_pnt
+ *
+ * DESCRIPTION:
+ *
+ * This function is mainly used by the arg_check.c functions to verify
+ * specific pointers.  This can be used by users to provide more fine
+ * grained tests on pointers.
+ *
+ * RETURNS:
+ *
+ * Success - MALLOC_VERIFY_NOERROR
+ *
+ * Failure - MALLOC_VERIFY_ERROR
+ *
+ * ARGUMENTS:
+ *
+ * file -> File-name or return-address of the caller.  You can use
+ * __FILE__ for this argument or 0L for none.
+ *
+ * line -> Line-number of the caller.  You can use __LINE__ for this
+ * argument or 0 for none.
+ *
+ * func -> Function string which is checking the pointer.  0L if none.
+ *
+ * pnt -> Pointer we are checking.
+ *
+ * exact_b -> Set to 1 if this pointer was definitely handed back from
+ * a memory allocation.  If set to 0 then this pointer can be inside
+ * another allocation or outside the heap altogether.
+ *
+ * min_size -> Make sure that pointer can hold at least that many
+ * bytes if inside of the heap.  If -1 then make sure it can handle
+ * strlen(pnt) + 1 bytes (+1 for the \0).  If 0 then don't check the
+ * size.
+ */
+int	dmalloc_verify_pnt(const char *file, const int line, const char *func,
+			   const void *pnt, const int exact_b,
+			   const int min_size)
+{
+  int	ret;
+  
+  if (! dmalloc_in(file, line, 0)) {
+    return MALLOC_VERIFY_NOERROR;
+  }
+  
+  /* call the pnt checking chunk code */
+  ret = _dmalloc_chunk_pnt_check(func, pnt, exact_b, min_size);
+  dmalloc_out();
+  
+  if (ret) {
+    return MALLOC_VERIFY_NOERROR;
+  }
+  else {
+    return MALLOC_VERIFY_ERROR;
+  }
+}
+
 /*
  * unsigned int dmalloc_debug
  *
@@ -1563,6 +1652,11 @@ void	dmalloc_track(const dmalloc_track_t track_func)
  * dmalloc_log_changed to log the changed pointers since this point.
  * Multiple marks can be saved and used.
  *
+ * This is also the iteration number and can be logged at the front of
+ * each memory transaction in the logfile with the LOG_ITERATION
+ * define in settings.h and can be logged with each pointer with the
+ * LOG_PNT_ITERATION define in settings.h.
+ *
  * RETURNS:
  *
  * Current mark value
@@ -1580,6 +1674,30 @@ unsigned long	dmalloc_mark(void)
   return _dmalloc_iter_c;
 }
 
+/*
+ * unsigned long dmalloc_memory_allocated
+ *
+ * DESCRIPTION:
+ *
+ * Return the total number of bytes allocated by the program so far.
+ *
+ * RETURNS:
+ *
+ * Total number of bytes allocated by the program so far.
+ *
+ * ARGUMENTS:
+ *
+ * None.
+ */
+unsigned long	dmalloc_memory_allocated(void)
+{
+  if (! enabled_b) {
+    (void)dmalloc_startup(NULL /* no options string */);
+  }
+  
+  return _dmalloc_alloc_total;
+}
+
 /*
  * unsigned int dmalloc_page_size
  *
@@ -1604,6 +1722,48 @@ unsigned int	dmalloc_page_size(void)
   return BLOCK_SIZE;
 }
 
+/*
+ * unsigned long dmalloc_count_changed
+ *
+ * DESCRIPTION:
+ *
+ * Count the changed memory bytes since a particular mark.
+ *
+ * RETURNS:
+ *
+ * Number of bytes since mark.
+ *
+ * ARGUMENTS:
+ *
+ * mark -> Sets the point from which to count the changed memory.  You
+ * can use dmalloc_mark to get the current mark value which can later
+ * be passed in here.  Pass in 0 to report on the unfreed memory since
+ * the program started.
+ *
+ * not_freed_b -> Set to 1 to count the new pointers that are non-freed.
+ *
+ * free_b -> Set to 1 to count the new pointers that are freed.
+ */
+unsigned long	dmalloc_count_changed(const unsigned long mark,
+				      const int not_freed_b, const int free_b)
+{
+  unsigned long	mem_count;
+  
+  if (! dmalloc_in(DMALLOC_DEFAULT_FILE, DMALLOC_DEFAULT_LINE, 1)) {
+    return 0;
+  }
+  
+  if (! BIT_IS_SET(_dmalloc_flags, DEBUG_LOG_TRANS)) {
+    dmalloc_message("counting the unfreed memory since mark %lu", mark);
+  }
+  
+  mem_count = _dmalloc_chunk_count_changed(mark, not_freed_b, free_b);
+  
+  dmalloc_out();
+  
+  return mem_count;
+}
+
 /*
  * void dmalloc_log_status
  *
